/*
 * growlnotifier.mm - A simple Qt interface to Growl
 *
 * Copyright (C) 2005-2011  Remko Troncon, Michail Pishchagin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * You can also redistribute and/or modify this program under the
 * terms of the Psi License, specified in the accompanied COPYING
 * file, as published by the Psi Project; either dated January 1st,
 * 2005, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


extern "C" {
	#include <CoreFoundation/CoreFoundation.h>
	#include <Growl/Growl.h>
}

#include <QStringList>
#include <QPixmap>
#include <QBuffer>

#include "growlnotifier.h"
#include "privateqt_mac.h"

NSMutableArray* QStringListToNSMutableArray(const QStringList& stringList)
{
	NSMutableArray* result = [[NSMutableArray alloc] initWithCapacity:stringList.count()];
	foreach(const QString& i, stringList) {
		[result addObject:(NSString*)QtCFString::toCFStringRef(i)];
	}
	return result;
}

@interface QGrowlDelegate : NSObject<GrowlApplicationBridgeDelegate> {
	QStringList notifications;
	QStringList defaults;
}
@end

@implementation QGrowlDelegate
- (id) initWithNotifications:(const QStringList&)_notifications andDefaults:(const QStringList&)_defaults
{
	if((self = [super init])) {
		notifications = _notifications;
		defaults = _defaults;
	}
	return self;
}

- (NSDictionary*) registrationDictionaryForGrowl
{
	NSDictionary* growlReg = [NSDictionary dictionaryWithObjectsAndKeys:
		QStringListToNSMutableArray(notifications), GROWL_NOTIFICATIONS_ALL,
		QStringListToNSMutableArray(defaults), GROWL_NOTIFICATIONS_DEFAULT,
		nil];

	return growlReg;
}

- (void) performCallback:(NSString*)slot_name withContext:(id)clickContext
{
	if (clickContext && [clickContext isKindOfClass:[NSDictionary class]]) {
		QObject* receiver = 0;
		char* slot = 0;
		void* context = 0;

		CFDataRef data;
		data = (CFDataRef)[clickContext objectForKey:@"receiver"];
		CFDataGetBytes(data, CFRangeMake(0, CFDataGetLength(data)), (UInt8*)&receiver);

		data = (CFDataRef)[clickContext objectForKey:slot_name];
		CFDataGetBytes(data, CFRangeMake(0, CFDataGetLength(data)), (UInt8*)&slot);

		data = (CFDataRef)[clickContext objectForKey:@"context"];
		CFDataGetBytes(data, CFRangeMake(0, CFDataGetLength(data)), (UInt8*)&context);

		QMetaObject::invokeMethod(receiver, slot, Qt::QueuedConnection, QGenericReturnArgument(), Q_ARG(void*, context));
	}
}

- (void) growlNotificationWasClicked:(id)clickContext
{
	[self performCallback:@"clicked_slot" withContext:clickContext];
}

- (void) growlNotificationTimedOut:(id)clickContext
{
	[self performCallback:@"timeout_slot" withContext:clickContext];
}
@end

class GrowlNotifier::Private {
public:
	QtMacCocoaAutoReleasePool pool_;
	QGrowlDelegate* growlDelegate_;
};

GrowlNotifier::GrowlNotifier(QObject* parent, const QStringList& notifications, const QStringList& defaults, const QString& appName)
	: QObject(parent)
	, d(new Private())
{
	Q_UNUSED(appName);

	d->growlDelegate_ = [[QGrowlDelegate alloc] initWithNotifications:notifications andDefaults:defaults];
	[GrowlApplicationBridge setGrowlDelegate:d->growlDelegate_];
}

GrowlNotifier::~GrowlNotifier()
{
	[d->growlDelegate_ release];
	delete d;
}

void GrowlNotifier::notify(const QString& name, const QString& title,
	const QString& description, const QPixmap& p, bool sticky,
	const QObject* receiver,
	const char* clicked_slot, const char* timeout_slot,
	void* qcontext)
{
	CFDataRef icon = 0;
	if (!p.isNull()) {
		QByteArray imgData;
		QBuffer buffer(&imgData);
		buffer.open(QIODevice::WriteOnly);
		p.save(&buffer, "PNG");
		icon = CFDataCreate(NULL, (UInt8*)imgData.data(), imgData.size());
		[(id)icon autorelease];
	}

	NSDictionary* clickContext = [NSDictionary dictionaryWithObjectsAndKeys:
		[(id)CFDataCreate(kCFAllocatorDefault, (const UInt8*)&receiver, sizeof(const QObject*)) autorelease], @"receiver",
		[(id)CFDataCreate(kCFAllocatorDefault, (const UInt8*)&clicked_slot, sizeof(const char*)) autorelease], @"clicked_slot",
		[(id)CFDataCreate(kCFAllocatorDefault, (const UInt8*)&timeout_slot, sizeof(const char*)) autorelease], @"timeout_slot",
		[(id)CFDataCreate(kCFAllocatorDefault, (const UInt8*)&qcontext, sizeof(void*)) autorelease], @"context",
		nil];

	[GrowlApplicationBridge notifyWithTitle:[(NSString*)QtCFString::toCFStringRef(title) autorelease]
	                            description:[(NSString*)QtCFString::toCFStringRef(description) autorelease]
	                       notificationName:[(NSString*)QtCFString::toCFStringRef(name) autorelease]
	                               iconData:(NSData*)icon
	                               priority:0
	                               isSticky:sticky
	                           clickContext:clickContext];
}
