/*
 * yandexspeller.cpp
 * Copyright (C) 2009  Michail Pishchagin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "yandexspeller.h"

#include <QNetworkAccessManager>
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QVariant>
#include <QDomDocument>
#include <QTimer>
#include <QMutableListIterator>
#include <QSettings>

#ifdef YAPSI
#include "psioptions.h"
static const QString customDictionaryOptionPath = "options.ui.spell-check.custom-dictionary";
#endif

#include "httphelper.h"

static const int MAX_WORDS_PER_REQUEST = 100;
#define USE_HTTPS

YandexSpeller::YandexSpeller(PsiCon* controller)
	: BaseHttpHelper<SpellChecker>(controller)
{
	spellingSuggestions_.setMaxCost(100000);

	processRequestsTimer_ = new QTimer(this);
	processRequestsTimer_->setSingleShot(true);
	processRequestsTimer_->setInterval(500);
	connect(processRequestsTimer_, SIGNAL(timeout()), SLOT(processRequests()));

	saveCustomDictionaryTimer_ = new QTimer(this);
	saveCustomDictionaryTimer_->setSingleShot(true);
	saveCustomDictionaryTimer_->setInterval(1000);
	connect(saveCustomDictionaryTimer_, SIGNAL(timeout()), SLOT(saveCustomDictionary()));

	customDictionary_.clear();
	QStringList stringList;
#ifdef YAPSI
	stringList = PsiOptions::instance()->getOption(customDictionaryOptionPath).toStringList();
#else
	QSettings settings("Yandex", "QtSpeller");
	stringList = settings.value("customDictionary", QStringList()).toStringList();
#endif
	foreach(const QString& word, stringList) {
		customDictionary_[word] = true;
	}
}

YandexSpeller::~YandexSpeller()
{
#ifndef YAPSI
	saveCustomDictionary();
#endif
}

void YandexSpeller::saveCustomDictionary()
{
	QStringList words;
	QHashIterator<QString, bool> it(customDictionary_);
	while (it.hasNext()) {
		it.next();
		words << it.key();
	}
#ifdef YAPSI
	PsiOptions::instance()->setOption(customDictionaryOptionPath, words);
#else
	QSettings settings("Yandex", "QtSpeller");
	settings.setValue("customDictionary", words);
#endif
}

bool YandexSpeller::available() const
{
	return true;
}

bool YandexSpeller::writable() const
{
	return true;
}

QList<QString> YandexSpeller::suggestions(const QString& word)
{
	if (spellingSuggestions_.contains(word)) {
		QStringList* list = spellingSuggestions_[word];
		if (list) {
			return *list;
		}
	}
	return QList<QString>();
}

bool YandexSpeller::isSpeltCorrectly(const QString& word, SyntaxHighlighter* highlighter)
{
	Q_UNUSED(highlighter);
	return customDictionary_.contains(word) ||
	       !spellingSuggestions_.contains(word);
}

bool YandexSpeller::learnSpelling(const QString& word)
{
	customDictionary_[word] = true;
	saveCustomDictionaryTimer_->start();
	return true;
}

QList<QtTextRange> YandexSpeller::spellingErrorIndexes(const QString& text, SyntaxHighlighter* highlighter, int cursorPositionInCurrentBlock, bool* needRehighlight, int blockCount)
{
	Q_UNUSED(blockCount);
	Q_ASSERT(needRehighlight);

	QList<QtTextRange> ranges;

	QStringList wordsToCheck;

	foreach(const QtTextRange& range, splitTextIntoWords(text, needRehighlight, cursorPositionInCurrentBlock)) {
		QString word = text.mid(range.index, range.length);
		if (spellingSuggestions_.contains(word)) {
			if (!customDictionary_.contains(word) && spellingSuggestions_[word]) {
				*needRehighlight = true;
				ranges << range;
			}
		}
		else {
			*needRehighlight = true;
			if (wordsToCheck.length() < MAX_WORDS_PER_REQUEST) {
				wordsToCheck << word;
			}
		}
	}

	if (!wordsToCheck.isEmpty()) {
		QMutableListIterator<SpellingRequest> it(requestQueue_);
		while (it.hasNext()) {
			SpellingRequest r = it.next();
			if (r.highlighter == highlighter) {
				it.remove();
			}
		}

		requestQueue_ << SpellingRequest(wordsToCheck, highlighter);
		processRequestsTimer_->start();
		*needRehighlight = true;
	}

	return ranges;
}

void YandexSpeller::processRequests()
{
	QMutableListIterator<SpellingRequest> it(requestQueue_);
	while (it.hasNext()) {
		SpellingRequest request = it.next();
		if (request.reply)
			continue;

		QString scheme = "http";
#ifdef USE_HTTPS
		scheme = "https";
#endif
		const int BY_WORDS = 0x100;
		int options = BY_WORDS;
		request.reply = network()->get(
		                    HttpHelper::getRequest(
		                        QString("%3://speller.yandex.net/services/spellservice/checkText?options=%1&text=%2")
		                        .arg(options)
		                        .arg(request.words.join(" "))
		                        .arg(scheme)));
		it.setValue(request);
	}
}

void YandexSpeller::replyFinished(QNetworkReply* reply)
{
	reply->deleteLater();

	QMutableListIterator<SpellingRequest> it(requestQueue_);
	while (it.hasNext()) {
		SpellingRequest request = it.next();
		if (request.reply != reply)
			continue;

		it.remove();
		if (reply->error() != QNetworkReply::NoError) {
			qWarning("YandexSpeller::replyFinished(): error");
			break;
		}

		QByteArray data = reply->readAll();
		QString text = QString::fromUtf8(data);

		QDomDocument doc;
		if (doc.setContent(data)) {
			QDomElement root = doc.documentElement();
			if (root.tagName() == "SpellResult") {
				QStringList correctWords = request.words;
				for (QDomNode n = root.firstChild(); !n.isNull(); n = n.nextSibling()) {
					QDomElement e = n.toElement();
					if (e.isNull() || e.tagName() != "error")
						continue;

					QString word;
					QStringList* suggestions = new QStringList();
					for (QDomNode n2 = e.firstChild(); !n2.isNull(); n2 = n2.nextSibling()) {
						QDomElement e2 = n2.toElement();
						if (e2.isNull())
							continue;
						if (e2.tagName() == "word")
							word = e2.text();
						if (e2.tagName() == "s")
							suggestions->append(e2.text());
					}
					correctWords.removeAll(word);
					spellingSuggestions_.insert(word, suggestions);
				}

				foreach(const QString& word, correctWords) {
					spellingSuggestions_.insert(word, 0);
				}

				if (request.highlighter) {
					request.highlighter->rehighlight();
				}
			}
		}
		break;
	}
}
